local MAJOR_VERSION = "LibShieldLeft-1.0"
local MINOR_VERSION = tonumber(("$Revision: 5 $"):match("%d+"))

local lib = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local playerName = UnitName('player')
local playerClass = select(2, UnitClass('player'))
local isShielder = (playerClass == "PRIEST") 


-----------------
-- Event Frame --
-----------------

lib.EventFrame = lib.EventFrame or CreateFrame("Frame")
lib.EventFrame:SetScript("OnEvent", function (this, event, ...) lib[event](lib, ...) end)
lib.EventFrame:UnregisterAllEvents()

-- Register Events
lib.EventFrame:RegisterEvent("LEARNED_SPELL_IN_TAB")
lib.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

-- Prune data at zone change
lib.EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
lib.EventFrame:RegisterEvent("RAID_ROSTER_UPDATE")
lib.EventFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")


----------------------
-- Scanning Tooltip --
----------------------

if (not lib.Tooltip) then
    lib.Tooltip = CreateFrame("GameTooltip")
    lib.Tooltip:SetOwner(UIParent, "ANCHOR_NONE")
    for i = 1, 4 do
        lib["TooltipTextLeft" .. i] = lib.Tooltip:CreateFontString()
        lib["TooltipTextRight" .. i] = lib.Tooltip:CreateFontString()
        lib.Tooltip:AddFontStrings(lib["TooltipTextLeft" .. i], lib["TooltipTextRight" .. i])
    end
end


-------------------------------
-- Embed CallbackHandler-1.0 --
-------------------------------

lib.Callbacks = LibStub("CallbackHandler-1.0"):New(lib)

-----------------
-- Static Data --
-----------------

-- Cache of spells and shield sizes
local SpellCache = {}

local Shields = {
    --Power World: Shield
    [GetSpellInfo(17)] = {
        name = GetSpellInfo(17),
        duration = 30,
        order = 30,
        minAmount = {
            [12] = 1265,
            [13] = 1920,
            [14] = 2230,
        },
    },
    --Divine Aegis
    [GetSpellInfo(47753)] = {
        name = GetSpellInfo(47753),
        duration = 12,
        order = 12,
    },
    --Ice Barrier
    [GetSpellInfo(13032)] = {
        name = GetSpellInfo(13032),
        duration = 60,
        order = 100,
        minAmount = {
            [6] = 1075,
            [7] = 2800,
            [8] = 3300,
        },
    },
    --Mana Shield
    [GetSpellInfo(10193)] = {
        name = GetSpellInfo(10193),
        duration = 60,
        order = 150,
        minAmount = {
            [7] = 715,
            [8] = 1080,
            [9] = 1330,
        }
    },
    --Fire Ward
    [GetSpellInfo(43010)] = {
        name = GetSpellInfo(43010),
        duration = 30,
        order = 0,
        school = 0x04, --fire
        minAmount = {
            [6] = 1125,
            [7] = 1950,
        }
    },
    --Frost Ward
    [GetSpellInfo(6143)] = {
        name = GetSpellInfo(6143),
        duration = 30,
        order = 0,
        school = 0x10, --frost
        minAmount = {
            [6] = 1125,
            [7] = 1950,
        }
    },
    --Hand of protection
    [GetSpellInfo(5599)] = {
        name = GetSpellInfo(5599),
        duration = 8,
        order = 0,
    },
    --Sacred Shield
    [GetSpellInfo(53601)] = {
        name = GetSpellInfo(53601),
        duration = 6,
        order = 6,
        minAmount = 500,
    },
    --Shadow ward 6229
    [GetSpellInfo(6229)] = {
        name = GetSpellInfo(6229),
        duration = 30,
        order = 0,
        school = 0x20,
        minAmount = {
            [4] = 875,
            [5] = 2750,
            [6] = 3300,
        },
    },
    --Fell blossom 28527
    [GetSpellInfo(28527)] = {
        name = GetSpellInfo(28527),
        duration = 15,
        order = 15,
        minAmount = 750,
    },
}

local EventParse =
{
	
}

-- active Shields
local unitData = {}

local currentTimestamp = 0
local timeGone = 0
local aegisTolerance = 0.3

---------------------------------
-- Frequently Accessed Globals --
---------------------------------
local bit_band = _G.bit.band


---------------
-- Utilities --
---------------

local function unitFullName(unit)
    local name, realm = UnitName(unit)
    if (realm and realm ~= "") then
        return name .. "-" .. realm
    else
        return name
    end
end

-- Spellbook Scanner --
local function getBaseShieldSize(name)

    -- Check if info is already cached
    if (SpellCache[name]) then
        return SpellCache[name]
    end

    SpellCache[name] = {}

    -- Gather info (only done if not in cache)
    local i = 1

    while true do

        local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
        
        if (not spellName) then 
            break 
        end

        if (spellName == name) then
            -- This is the spell we're looking for, gather info

            -- Determine rank
            spellRank = tonumber(spellRank:match("(%d+)"))
            lib.Tooltip:SetSpell(i, BOOKTYPE_SPELL)
    
            -- Determine healing
            local shieldSize = select(3, string.find(lib.TooltipTextLeft4:GetText() or lib.TooltipTextLeft3:GetText() or "", "[(absorb)]+.-(%d+).-[(damage)(Schaden)]+"))
            
            SpellCache[spellName][spellRank] = shieldSize
        end
        i = i + 1
    end

    return SpellCache[name]
end


-- Detects if a buff is present on the unit and returns the application number
local function detectBuff(unit, buffName)
    local idTable = {}
    local buffID = 1
    local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(unit, buffID) 
    
    while name do
        if name == buffName then
            tinsert(idTable, buffID)
        end
        buffID = buffID + 1
        name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(unit, buffID) 
    end
    
    return idTable
end

local function isInPartyOrRaidOrMe(unitFlags)
    return  bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_RAID) == COMBATLOG_OBJECT_AFFILIATION_RAID or 
            bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY) == COMBATLOG_OBJECT_AFFILIATION_PARTY or 
            bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) == COMBATLOG_OBJECT_AFFILIATION_MINE 
end


function debug(msg)
    --print(msg)
end

-----------------------------
-- Unit Data Management --
-----------------------------

local function getUnitData(unitGUID, dontCreate)
    local data = unitData[unitGUID]
    if not data and not dontCreate then
        data = {}
        unitData[unitGUID] = data
    end
    
    return data
end

local function getUnitActiveShields(unitGUID, dontCreate)
    local data = getUnitData(unitGUID, dontCreate)
    
    if not data then return end
    
    local activeShields = data.activeShields
        
    if not activeShields and not dontCreate then
        activeShields = {amount = 0, count = 0}
        data.activeShields = activeShields
    end
    
    return activeShields
end

----------------------
-- Public Functions --
----------------------

function lib:SetAegisTolerance(tolerance)
    if tolerance then
        aegisTolerance = tolerance
    end
end

--------------------
-- Class Specific --
--------------------

local ClassSpecific = nil


-- Priest --

if (playerClass == "PRIEST") then
    ClassSpecific = {}
    
    local PWShield = GetSpellInfo(17)
    local DivineAegis = GetSpellInfo(47753)

    ClassSpecific.Heal = function (lib, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, amount, overhealing, critical)
        if not srcName then
            print("Error: Got a heal ("..spellName..") withouth source on "..dstName)
            return
        end
        if critical and srcName == playerName then
            debug(timeGone..": Crit Heal")
            local data = getUnitData(dstGUID)  

            if data.lastCritTime and data.lastAegis then
                local critBeforeAegis = data.lastAegis - data.lastCritTime
                local critAfterAegis = currentTimestamp - data.lastAegis
                
                data.lastCritAmount = amount            
                data.lastCritTime = currentTimestamp
                data.lastAegis = nil
                --check if the crit after the aegis proc is way nearer than the crit before it
                if critAfterAegis <= aegisTolerance * 0.5 and critAfterAegis < 0.5 * critBeforeAegis then
                    debug("Crit short after aegis, updating.")
                    lib:RefreshShield(Shields[DivineAegis], dstGUID, unitFullName(dstName))
                end
            else
                data.lastCritAmount = amount            
                data.lastCritTime = currentTimestamp
            end
        end
    end

    EventParse["SPELL_HEAL"] = ClassSpecific.Heal
    EventParse["SPELL_PERIODIC_HEAL"] = ClassSpecific.Heal
    EventParse["SPELL_BUILDING_HEAL"] = ClassSpecific.Heal
    EventParse["RANGE_HEAL"] = ClassSpecific.Heal
    
    ClassSpecific.GetDurationForShield = function(shieldData, target)        
        return shieldData.duration
    end

    ClassSpecific.GatherShieldData = function(shieldData, tgtGUID, target, rank, duration, count)
        
        if shieldData == Shields[PWShield] then            
            local base = getBaseShieldSize(shieldData.name)[rank]
            if not base then 
                print("Error: Could not find base shield size for "..shieldData.name)
                return lib:DefaultShieldData(shieldData, rank, duration, count)
            end
            
            local bonus = GetSpellBonusHealing();
            
            -- Twin disciplines - increases heal/dmg of instants by 1/2/3/4/5% 
            local _, _, _, _, talentTwinDisc = GetTalentInfo(1,2);   
            talentTwinDisc = 1 + talentTwinDisc * 0.01

            -- Borrowed Time - increases absorbed value by 8/16/24/32/40% of spell power
            local _, _, _, _, talentBrwdTime = GetTalentInfo(1,27);   
            talentBrwdTime = talentBrwdTime * 0.08
            
            -- Improved Power word: Shield - increases absorbed value by 5/10/15%
            local _, _, _, _, talentImprShield = GetTalentInfo(1,7);     
            talentImprShield = 1 + talentImprShield * 0.05
            
            -- Focused Power: - increases heal/dmg by 2/4%
            local _, _, _, _, talentFocusedPower = GetTalentInfo(1,16);     
            talentFocusedPower = 1 + talentFocusedPower * 0.02
            
            --return ((base + bonus * 0.8057) * talentFocusedPower + bonus * talentBrwdTime) * talentImprShield * talentTwinDisc
            return (base + bonus * (0.42857*1.88 + talentBrwdTime)) * talentImprShield * talentTwinDisc * talentFocusedPower
            
        elseif shieldData == Shields[DivineAegis] then
            local data = getUnitData(tgtGUID)
            
            --in can happen that the crit heal message comes after the aegis message
            --this value is used at the next crit to check if this had happened
            data.lastAegis = currentTimestamp
            
            if not data.lastCritTime or currentTimestamp - data.lastCritTime > aegisTolerance then                
                debug("No crit before aegis")
                return lib:DefaultShieldData(shieldData, rank, duration, count)
            end
            
            -- Divine Aegis - adds an aegis which absorbes 10/20/30% of last crit heal
            local _, _, _, _, talentDivAegis = GetTalentInfo(1,24);  
            if not talentDivAegis or talentDivAegis == 0 then 
                print("Error: Aegis of player although he doesn't have skilled it. Please report this error to addon author.")   
                return lib:DefaultShieldData(shieldData, rank, duration, count)
            end
            
            return data.lastCritAmount * talentDivAegis * 0.1, duration
        else
            print("Error: Not handled shield("..shieldData.name.."! Please report this error to addon author.")
            return lib:DefaultShieldData(shieldData, rank, duration, count)
        end
    end
end

function lib:DefaultShieldData(shieldData, rank, duration, count)
    if not shieldData.minAmount then
        return 0, duration
    end
    
    if type(shieldData.minAmount) == "table" then
        return shieldData.minAmount[rank], duration
    else
        return shieldData.minAmount, duration
    end
end

function lib:GatherShieldData(shieldData, tgtGUID, target)
    local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(target, shieldData.name)
        
    if not name then
        print("Error: Could not find "..shieldData.name.." as aura on unit!")
        return 0, 0
    end
    
    rank = tonumber(rank:match("(%d+)"))
    
    if isMine and ClassSpecific then
        return ClassSpecific.GatherShieldData(shieldData, tgtGUID, target, rank, duration, count)
    else
        return self:DefaultShieldData(shieldData, rank, duration, count)
    end
end

function lib:RemoveShieldFromList(shieldData, list)
    local shield = list.first
    
    if not shield then
        return
    end
    
    if shield.data == shieldData then --shield to be removed is first
        list.first = shield.next
        
    else
        --find predecessor of the shield to be removed
        while shield.next and shield.next.data ~= shieldData and shield.next.data.order <= shieldData.order do
            shield = shield.next
        end
        
        if not shield.next or shield.next.data ~= shieldData then --shield not in list
            return
        end
        
        --remove shield from list
        local buff = shield.next
        shield.next = buff.next
        shield = buff
    end
    
    return shield
end

function lib:GetShieldFromList(shieldData, list)
    --get head of shield list
    local shield = list.first
    
    if not shield  or shield.data.order > shieldData.order then --shield list is empty or our new entry is smallest, add new shield as first        
       return nil
    else --otherwise find position to insert new shieldData entry
        while shield.data ~= shieldData and shield.next and shield.data.order <= shieldData.order do
            shield = shield.next
        end
        
        if shield.data ~= shieldData then
            return nil
        end
    end
    
    return shield
end

function lib:NewShieldToList(shieldData, list)
    local shield = list.first
    
    if not shield  or shield.data.order > shieldData.order then --shield list is empty or our new entry is smallest, add new shield as first
        shield = {data = shieldData} --create new entry
        shield.next = list.first --set current first as next of new entry
        list.first = shield --new entry becomes new current first
    else --otherwise find position to insert new shieldData entry
        while shield.data ~= shieldData and shield.next and shield.data.order <= shieldData.order do
            shield = shield.next
        end
        
        if shield.data ~= shieldData then
            local buff = shield.next --buffer successor
            shield.next = {data = shieldData} --create new entry as successor
            shield = shield.next --go to new entry
            shield.next = buff --set buffer as next of new entry
        else
            return shield, true
        end
    end
    
    return shield
end

function lib:NewShield(shieldData, unitGUID, unitName)
    debug(timeGone..": New shield("..shieldData.name..") on "..unitName)
    --get active shields of the unit
    local activeShields = getUnitActiveShields(unitGUID)
    
    local shield, wasInList = self:NewShieldToList(shieldData, activeShields)
    
    if wasInList then
        print("Error: New shield already in list")
        activeShields.count = activeShields.count - 1
        if shield.amountLeft then
            activeShields.amount = activeShields.amount - shield.amountLeft
        end
    end
    
    shield.amountLeft, shield.expTime = self:GatherShieldData(shieldData, unitGUID, unitName)
    debug(shield.amountLeft)
    if shield.amountLeft > 0 then
        activeShields.amount = (activeShields.amount or 0) + shield.amountLeft
    end
    
    activeShields.count = activeShields.count + 1
    self.Callbacks:Fire("ShieldLeft_NewShield", unitGUID, unitName, shieldData.name, shield.amountLeft, activeShields.amount, activeShields.count)
end

function lib:RefreshShield(shieldData, unitGUID, unitName, expTime)
    debug(timeGone..": Refresh shield("..shieldData.name..") on "..unitName)
    --get active shields of the unit
    local activeShields = getUnitActiveShields(unitGUID)
    
    --get head of shield list
    local shield = self:GetShieldFromList(shieldData, activeShields)
    
    if not shield then
        print("Error: Refreshed shield not in list")
        shield = self:NewShieldToList(shieldData, activeShields)
        shield.amountLeft = 0
        shield.expTime = 0
        activeShields.count = activeShields.count + 1
    end
    
    local newAmount
    newAmount, shield.expTime, additive = self:GatherShieldData(shieldData, unitGUID, unitName, true)
    debug(newAmount)
    if newAmount > 0 then
        if additive then
            activeShields.amount = (activeShields.amount or 0) + newAmount
            shield.amountLeft = (shield.amountLeft or 0) + newAmount
        else
            activeShields.amount = activeShields.amount + (newAmount - shield.amountLeft)
            shield.amountLeft = newAmount
        end
    end
    self.Callbacks:Fire("ShieldLeft_RefreshShield", unitGUID, unitName, shieldData.name, shield.amountLeft, activeShields.amount, activeShields.count)
end

function lib:RemoveShield(shieldData, unitGUID, unitName)
    debug(timeGone..": Remove shield("..shieldData.name..") on "..unitName)
    --get active shields of the unit
    local activeShields = getUnitActiveShields(unitGUID, true)
    if not activeShields then
        return
    end
    
    --get head of shield list
    local shield = self:RemoveShieldFromList(shieldData, activeShields)
    
    if not shield then  --shield was not in list
        print("Error: Removed shield was not in list.")
        return
    end
    
    debug("Shield gone with "..shield.amountLeft.." left.")
    --a negative amount left means the shield has absorbed more than it could (or we expect)
    --if the last hit just happened we transfer this amount as damage to the next shield
    if shield.amountLeft < 0 then        
        if currentTimestamp - getUnitData(unitGUID).lastAbsorb < 1 then
            debug("Forwarding damage to next shield")
            self:UnitAbsorbed(unitName, -shield.amountLeft, shield.data.school or 0x01, false)
        end
        
    --a positive amoutn left means the shield has absorbed less than it could (or we expected)
    --this could happen because the actual absorb message of the damage which destroyed this shield
    --comes short after this message. In this case we forward the capacity left to the next shield
    --(because it has to absorb the damage this one actually had)
    elseif shield.amountLeft > 0 then
        if shield.next then
            getUnitData(unitGUID).lastRemove = currentTimestamp
            shield.next.forwardCap = shield.amountLeft
        end
        --since we don't know if in the next milliseconds comes the actual absorb message which destroys this
        --shield we have to assume the amount left lost
        activeShields.amount = activeShields.amount - shield.amountLeft
    end
    
    activeShields.count = activeShields.count - 1    
    self.Callbacks:Fire("ShieldLeft_RemoveShield", unitGUID, unitName, shieldData.name, shield.amountLeft, activeShields.amount, activeShields.count)
end

function lib:UnitAbsorbed(unitGUID, unitName, amount, school, partly)
    debug(timeGone..": Absorb on "..unitName.." for "..amount)
    --get active shields of the unit
    local activeShields = getUnitActiveShields(unitGUID, true)
    if not activeShields then
        --print("no active shield")
        return
    end
    
    --get head of shield list
    local shield = activeShields.first
    
    while shield and shield.data.school and shield.data.school ~= school do
        shield = shield.next
    end
    
    if not shield then  --shield list is empty or our new entry is smallest
        --print("no shield or no appliable")
        return          --shield is not in list
    end
    
    debug(timeGone..": Absorb shield("..shield.data.name..") on "..unitName.." for "..amount)
    
    local time = currentTimestamp
    --forwardCap is set if the last shield removed had some capacity left
    --if this absorb comes short after the remove it might be that
    --this is its actual absorb message which removed the shield
    if shield.forwardCap then
        local lastRemove = getUnitData(unitGUID).lastRemove
        if lastRemove and (time - lastRemove) < 0.5 then
            --there was a remove wis capacity left short before so add the capacity to this shield 
            --(but max the amount of this absorb message)
            local bonusCap = min(amount, shield.forwardCap)
            debug("Forwarding capacity of "..bonusCap.." to next shield")
            shield.amountLeft = shield.amountLeft + bonusCap
            activeShields.amount = activeShields.amount +  bonusCap
        end
        shield.forwardCap = nil
    end
    getUnitData(unitGUID).lastAbsorb = time
      
    --an negative amount left ist transfered as damage to the next shield 
    --but maximum the amount of the last hit
    if shield.amountLeft > 0 then
        shield.amountLeft = shield.amountLeft - amount
        activeShields.amount = activeShields.amount - (amount + min(shield.amountLeft,0))
    else
        shield.amountLeft = -amount
    end  
    
    self.Callbacks:Fire("ShieldLeft_UpdateShield", unitGUID, unitName, shield.data.name, shield.amountLeft, activeShields.amount, activeShields.count)   
end

function lib:CaptureRunningShields()
    if GetNumRaidMembers() > 0 then
        for id = 1, GetNumRaidMembers() do
            local unitid = "raid"..id
            local name = UnitName(unitid)
            if name then
                self:CaptureUnitsRunningShields(UnitGUID(unitid), name)
            end
        end
    elseif GetNumPartyMembers() > 0 then
        for id = 1, GetNumPartyMembers() do
            local unitid = "party"..id
            local name = UnitName(unitid)            
            if name then
                self:CaptureUnitsRunningShields(UnitGUID(unitid), name)
            end
        end
    end
    self:CaptureUnitsRunningShields(UnitGUID("player"), playerName)
end

function lib:CaptureUnitsRunningShields(unitGUID, unitName)
    debug("Capturing shields on "..unitName)
    local fullName = unitFullName(unitName)
    if not fullName then 
        print("Error: Could not get full name for unit:"..unitName)
        return 
    end
    
    --if getUnitData(fullName, true) then --dont capture for units we already know
    --    return
    --end
    
    local activeShields = getUnitActiveShields(unitGUID, true)
        
    local buffID = 1
    local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(unitName, buffID) 
    
    while name do
        debug("Found aura "..name)
        local shieldData = Shields[name]
        if shieldData then
            debug("Found shield aura "..shieldData.name)
            if activeShields then
                local shield = self:GetShieldFromList(shieldData, activeShields)
                if not shield then
                    debug("New shield "..shieldData.name.." after zone switch.")
                    self:NewShield(shieldData, unitGUID, fullName)   
                else
                    debug("Already in list")
                end
                shield = self:GetShieldFromList(shieldData, activeShields)
                shield.stillActive = true
            else
                self:NewShield(shieldData, unitGUID, fullName)
            end            
        end
        buffID = buffID + 1
        name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff(unitName, buffID) 
    end
    
    if activeShields then
        local shield = activeShields.first
    
        while shield do
            local tmp = shield
            shield = shield.next
            if not tmp.stillActive then
                debug("Shield "..tmp.data.name.." run out during zone switch.")
                self:RemoveShield(tmp.data, unitGUID, fullName)
            end            
        end
    end
end

--------------------
-- Event Handlers --
--------------------

function lib:AuraApplied(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    local shieldData = Shields[spellName]
    
    --debug("aura "..spellName..": "..spellId)
    
    if not shieldData then return end
    
    local fullName = unitFullName(dstName)
    if not fullName then return end
    
    self:NewShield(shieldData, dstGUID, fullName)
end

function lib:AuraRefresh(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    local shieldData = Shields[spellName]
    
    if not shieldData then return end
    
    local fullName = unitFullName(dstName)
    if not fullName then return end
    
    self:RefreshShield(shieldData, dstGUID, fullName)
end

function lib:AuraRemoved(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    local shieldData = Shields[spellName]
    
    if not shieldData then return end
    
    local fullName = unitFullName(dstName)
    if not fullName then return end
    
    self:RemoveShield(shieldData, dstGUID, fullName) 
end

function lib:AuraBroken(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    local shieldData = Shields[spellName]
    
    if not shieldData then return end
    debug("Shield broken")
    local fullName = unitFullName(dstName)
    if not fullName then return end
    
    
    self:RemoveShield(shieldData, dstGUID, fullName)  
end

function lib:AuraBrokenSpell(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    local shieldData = Shields[spellName]
    
    if not shieldData then return end
    debug("Shield broken")
    local fullName = unitFullName(dstName)
    if not fullName then return end
    
    
    self:RemoveShield(shieldData, dstGUID, fullName, true) 
end

function lib:SwingDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, amount, overkill, school, resisted, blocked, absorbed)  
    if absorbed and absorbed > 0 then
        local fullName = unitFullName(dstName)
        if not fullName then return end
    
        self:UnitAbsorbed(dstGUID, fullName, absorbed, 0x01, true)  
    end
end

function lib:SwingMissed(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, missType, amount)    
    if missType == "ABSORB" then
        local fullName = unitFullName(dstName)
        if not fullName then return end
        
        self:UnitAbsorbed(dstGUID, fullName, amount, 0x01, false)  
    end
end

function lib:SpellDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed)    
    if absorbed and absorbed > 0 then
        local fullName = unitFullName(dstName)
        if not fullName then return end
        
        self:UnitAbsorbed(dstGUID, fullName, absorbed, spellSchool, true)  
    end
end

function lib:SpellMissed(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, missType, amount)    
    if missType == "ABSORB" then
        local fullName = unitFullName(dstName)
        if not fullName then return end
        
        self:UnitAbsorbed(dstGUID, fullName, amount, spellSchool, false)  
    end
end

local envSchools = {
    
    ["FIRE"] = 0x04,
    ["LAVA"] = 0x04,
    ["SLIME"] = 0x20,
}

function lib:EnvironmentalDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, environmentalType, amount, overkill, school, resisted, blocked, absorbed)    
    if absorbed and absorbed > 0 then
        local fullName = unitFullName(dstName)
        if not fullName then return end
        
        local school = envSchools[environmentalType]
        if school then
            self:UnitAbsorbed(fullName, absorbed, school, true)  
        end
    end
end

function lib:EnvironmentalMissed(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, environmentalType, missType, amount)    
    if missType == "ABSORB" then
        local fullName = unitFullName(dstName)
        if not fullName then return end
        
        local school = envSchools[environmentalType]
        if school then
            self:UnitAbsorbed(fullName, absorbed, school, true)  
        end
    end
end

function lib:LEARNED_SPELL_IN_TAB()
    -- Invalidate cached spell data when learning new spells
    SpellCache = {};       
end

EventParse["SPELL_AURA_APPLIED"] = lib.AuraApplied
EventParse["SPELL_AURA_REFRESH"] = lib.AuraRefresh
EventParse["SPELL_AURA_REMOVED"] = lib.AuraRemoved 
EventParse["SPELL_AURA_BROKEN"] = lib.AuraBroken
EventParse["SPELL_AURA_BROKEN_SPELL"] = lib.AuraBrokenSpell

EventParse["SWING_DAMAGE"] = lib.SwingDamage
EventParse["SWING_MISSED"] = lib.SwingMissed
EventParse["ENVIRONMENTAL_DAMAGE"] = lib.EnvironmentalDamage
EventParse["ENVIRONMENTAL_MISSED"] = lib.EnvironmentalMissed
EventParse["SPELL_DAMAGE"] = lib.SpellDamage
EventParse["SPELL_MISSED"] = lib.SpellMissed
EventParse["SPELL_PERIODIC_DAMAGE"] = lib.SpellDamage
EventParse["SPELL_PERIODIC_MISSED"] = lib.SpellMissed
EventParse["SPELL_BUILDING_DAMAGE"] = lib.SpellDamage
EventParse["SPELL_BUILDING_MISSED"] = lib.SpellMissed
EventParse["RANGE_DAMAGE"] = lib.SpellDamage
EventParse["RANGE_MISSED"] = lib.SpellMissed



function lib:COMBAT_LOG_EVENT_UNFILTERED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)    
    if not isInPartyOrRaidOrMe(dstFlags) then
        return
    end
    
	local parsefunc = EventParse[eventtype]
	
	if parsefunc then  
        timeGone = timestamp - currentTimestamp
        currentTimestamp = timestamp
		parsefunc(self, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	end
end

function lib:PLAYER_ENTERING_WORLD()
    --clear cache
    self:CaptureRunningShields()
end

function lib:PARTY_MEMBERS_CHANGED()
    self:CaptureRunningShields()
end

function lib:RAID_ROSTER_UPDATE()
    self:CaptureRunningShields()
end
